Version 0.7.17
[userspace-rcu.git] / urcu-bp.c
index 0963dd05096d8ac5bc7f3c93520f064f75592a80..3e5be5fb59e1baf0ed7a7a7b7c26f61603654a98 100644 (file)
--- a/urcu-bp.c
+++ b/urcu-bp.c
@@ -79,8 +79,8 @@ void *mremap_wrapper(void *old_address, size_t old_size,
 }
 #endif
 
-/* Sleep delay in us */
-#define RCU_SLEEP_DELAY                1000
+/* Sleep delay in ms */
+#define RCU_SLEEP_DELAY_MS     10
 #define INIT_NR_THREADS                8
 #define ARENA_INIT_ALLOC               \
        sizeof(struct registry_chunk)   \
@@ -91,12 +91,29 @@ void *mremap_wrapper(void *old_address, size_t old_size,
  */
 #define RCU_QS_ACTIVE_ATTEMPTS 100
 
+static
+int rcu_bp_refcount;
+
 static
 void __attribute__((constructor)) rcu_bp_init(void);
 static
-void __attribute__((destructor)) rcu_bp_exit(void);
+void __attribute__((destructor)) _rcu_bp_exit(void);
 
+/*
+ * rcu_gp_lock ensures mutual exclusion between threads calling
+ * synchronize_rcu().
+ */
 static pthread_mutex_t rcu_gp_lock = PTHREAD_MUTEX_INITIALIZER;
+/*
+ * rcu_registry_lock ensures mutual exclusion between threads
+ * registering and unregistering themselves to/from the registry, and
+ * with threads reading that registry from synchronize_rcu(). However,
+ * this lock is not held all the way through the completion of awaiting
+ * for the grace period. It is sporadically released between iterations
+ * on the registry.
+ * rcu_registry_lock may nest inside rcu_gp_lock.
+ */
+static pthread_mutex_t rcu_registry_lock = PTHREAD_MUTEX_INITIALIZER;
 
 static pthread_mutex_t init_lock = PTHREAD_MUTEX_INITIALIZER;
 static int initialized;
@@ -168,10 +185,14 @@ static void mutex_unlock(pthread_mutex_t *mutex)
                urcu_die(ret);
 }
 
+/*
+ * Always called with rcu_registry lock held. Releases this lock between
+ * iterations and grabs it again. Holds the lock when it returns.
+ */
 void update_counter_and_wait(void)
 {
        CDS_LIST_HEAD(qsreaders);
-       int wait_loops = 0;
+       unsigned int wait_loops = 0;
        struct rcu_reader *index, *tmp;
 
        /* Switch parity: 0 -> 1, 1 -> 0 */
@@ -195,7 +216,9 @@ void update_counter_and_wait(void)
         * Wait for each thread rcu_reader.ctr count to become 0.
         */
        for (;;) {
-               wait_loops++;
+               if (wait_loops < RCU_QS_ACTIVE_ATTEMPTS)
+                       wait_loops++;
+
                cds_list_for_each_entry_safe(index, tmp, &registry, node) {
                        if (!rcu_old_gp_ongoing(&index->ctr))
                                cds_list_move(&index->node, &qsreaders);
@@ -204,10 +227,14 @@ void update_counter_and_wait(void)
                if (cds_list_empty(&registry)) {
                        break;
                } else {
-                       if (wait_loops == RCU_QS_ACTIVE_ATTEMPTS)
-                               usleep(RCU_SLEEP_DELAY);
+                       /* Temporarily unlock the registry lock. */
+                       mutex_unlock(&rcu_registry_lock);
+                       if (wait_loops >= RCU_QS_ACTIVE_ATTEMPTS)
+                               (void) poll(NULL, 0, RCU_SLEEP_DELAY_MS);
                        else
                                caa_cpu_relax();
+                       /* Re-lock the registry lock before the next loop. */
+                       mutex_lock(&rcu_registry_lock);
                }
        }
        /* put back the reader list in the registry */
@@ -225,6 +252,7 @@ void synchronize_rcu(void)
        assert(!ret);
 
        mutex_lock(&rcu_gp_lock);
+       mutex_lock(&rcu_registry_lock);
 
        if (cds_list_empty(&registry))
                goto out;
@@ -236,6 +264,8 @@ void synchronize_rcu(void)
 
        /*
         * Wait for previous parity to be empty of readers.
+        * update_counter_and_wait() can release and grab again
+        * rcu_registry_lock interally.
         */
        update_counter_and_wait();      /* 0 -> 1, wait readers in parity 0 */
 
@@ -248,6 +278,8 @@ void synchronize_rcu(void)
 
        /*
         * Wait for previous parity to be empty of readers.
+        * update_counter_and_wait() can release and grab again
+        * rcu_registry_lock interally.
         */
        update_counter_and_wait();      /* 1 -> 0, wait readers in parity 1 */
 
@@ -257,6 +289,7 @@ void synchronize_rcu(void)
         */
        cmm_smp_mb();
 out:
+       mutex_unlock(&rcu_registry_lock);
        mutex_unlock(&rcu_gp_lock);
        ret = pthread_sigmask(SIG_SETMASK, &oldmask, NULL);
        assert(!ret);
@@ -429,11 +462,8 @@ struct registry_chunk *find_chunk(struct rcu_reader *rcu_reader_reg)
 
 /* Called with signals off and mutex locked */
 static
-void remove_thread(void)
+void remove_thread(struct rcu_reader *rcu_reader_reg)
 {
-       struct rcu_reader *rcu_reader_reg;
-
-       rcu_reader_reg = URCU_TLS(rcu_reader);
        cleanup_thread(find_chunk(rcu_reader_reg), rcu_reader_reg);
        URCU_TLS(rcu_reader) = NULL;
 }
@@ -463,9 +493,9 @@ void rcu_bp_register(void)
         */
        rcu_bp_init();
 
-       mutex_lock(&rcu_gp_lock);
+       mutex_lock(&rcu_registry_lock);
        add_thread();
-       mutex_unlock(&rcu_gp_lock);
+       mutex_unlock(&rcu_registry_lock);
 end:
        ret = pthread_sigmask(SIG_SETMASK, &oldmask, NULL);
        if (ret)
@@ -474,7 +504,7 @@ end:
 
 /* Disable signals, take mutex, remove from registry */
 static
-void rcu_bp_unregister(void)
+void rcu_bp_unregister(struct rcu_reader *rcu_reader_reg)
 {
        sigset_t newmask, oldmask;
        int ret;
@@ -486,12 +516,13 @@ void rcu_bp_unregister(void)
        if (ret)
                abort();
 
-       mutex_lock(&rcu_gp_lock);
-       remove_thread();
-       mutex_unlock(&rcu_gp_lock);
+       mutex_lock(&rcu_registry_lock);
+       remove_thread(rcu_reader_reg);
+       mutex_unlock(&rcu_registry_lock);
        ret = pthread_sigmask(SIG_SETMASK, &oldmask, NULL);
        if (ret)
                abort();
+       _rcu_bp_exit();
 }
 
 /*
@@ -501,15 +532,14 @@ void rcu_bp_unregister(void)
 static
 void urcu_bp_thread_exit_notifier(void *rcu_key)
 {
-       assert(rcu_key == URCU_TLS(rcu_reader));
-       rcu_bp_unregister();
+       rcu_bp_unregister(rcu_key);
 }
 
 static
 void rcu_bp_init(void)
 {
        mutex_lock(&init_lock);
-       if (!initialized) {
+       if (!rcu_bp_refcount++) {
                int ret;
 
                ret = pthread_key_create(&urcu_bp_key,
@@ -522,24 +552,39 @@ void rcu_bp_init(void)
 }
 
 static
-void rcu_bp_exit(void)
+void _rcu_bp_exit(void)
 {
-       struct registry_chunk *chunk, *tmp;
-       int ret;
+       mutex_lock(&init_lock);
+       if (!--rcu_bp_refcount) {
+               struct registry_chunk *chunk, *tmp;
+               int ret;
 
-       cds_list_for_each_entry_safe(chunk, tmp,
-                       &registry_arena.chunk_list, node) {
-               munmap(chunk, chunk->data_len + sizeof(struct registry_chunk));
+               cds_list_for_each_entry_safe(chunk, tmp,
+                               &registry_arena.chunk_list, node) {
+                       munmap(chunk, chunk->data_len
+                                       + sizeof(struct registry_chunk));
+               }
+               ret = pthread_key_delete(urcu_bp_key);
+               if (ret)
+                       abort();
        }
-       ret = pthread_key_delete(urcu_bp_key);
-       if (ret)
-               abort();
+       mutex_unlock(&init_lock);
+}
+
+/*
+ * Keep ABI compability within stable versions. This has never been
+ * exposed through a header, but needs to stay in the .so until the
+ * soname is bumped.
+ */
+void rcu_bp_exit(void)
+{
 }
 
 /*
* Holding the rcu_gp_lock across fork will make sure we fork() don't race with
* a concurrent thread executing with this same lock held. This ensures that the
* registry is in a coherent state in the child.
  */
 void rcu_bp_before_fork(void)
 {
@@ -551,6 +596,7 @@ void rcu_bp_before_fork(void)
        ret = pthread_sigmask(SIG_BLOCK, &newmask, &oldmask);
        assert(!ret);
        mutex_lock(&rcu_gp_lock);
+       mutex_lock(&rcu_registry_lock);
        saved_fork_signal_mask = oldmask;
 }
 
@@ -560,6 +606,7 @@ void rcu_bp_after_fork_parent(void)
        int ret;
 
        oldmask = saved_fork_signal_mask;
+       mutex_unlock(&rcu_registry_lock);
        mutex_unlock(&rcu_gp_lock);
        ret = pthread_sigmask(SIG_SETMASK, &oldmask, NULL);
        assert(!ret);
@@ -567,7 +614,7 @@ void rcu_bp_after_fork_parent(void)
 
 /*
  * Prune all entries from registry except our own thread. Fits the Linux
* fork behavior. Called with rcu_gp_lock held.
  */
 static
 void urcu_bp_prune_registry(void)
@@ -595,6 +642,7 @@ void rcu_bp_after_fork_child(void)
 
        urcu_bp_prune_registry();
        oldmask = saved_fork_signal_mask;
+       mutex_unlock(&rcu_registry_lock);
        mutex_unlock(&rcu_gp_lock);
        ret = pthread_sigmask(SIG_SETMASK, &oldmask, NULL);
        assert(!ret);
@@ -522,24 +552,39 @@ void rcu_bp_init(void)
 
 /*
- * Holding the rcu_gp_lock across fork will make sure we fork() don't race with
- * a concurrent thread executing with this same lock held. This ensures that the
- * registry is in a coherent state in the child.
+ * Holding the rcu_gp_lock and rcu_registry_lock across fork will make
+ * sure we fork() don't race with a concurrent thread executing with
+ * any of those locks held. This ensures that the registry and data
+ * protected by rcu_gp_lock are in a coherent state in the child.
  */
 void rcu_bp_before_fork(void)
@@ -567,7 +614,7 @@ void rcu_bp_after_fork_parent(void)
 /*
  * Prune all entries from registry except our own thread. Fits the Linux
- * fork behavior. Called with rcu_gp_lock held.
+ * fork behavior. Called with rcu_gp_lock and rcu_registry_lock held.
  */
 static
This page took 0.025934 seconds and 4 git commands to generate.